from threading import Thread
from flask import current_app, render_template
from flask_mail import Mail, Message
from decorators import asyncfn

mail = Mail()

@asyncfn
def send_async_email(app, msg):
    with app.app_context():
        mail.send(msg)

def send_email(to, subject, template, **kwargs):
    if len(subject) > 50:
        newsubject = subject[:50] + '...'
    else:
        newsubject = subject
    app = current_app._get_current_object()
    msg = Message(app.config['MAIL_SUBJECT_PREFIX'] + ' ' + newsubject, sender=app.config['MAIL_SENDER'], recipients=[to])
    msg.body = render_template(template + '.txt', **kwargs)
    msg.html = render_template(template + '.html', **kwargs)
    thr = Thread(target=send_async_email, args=[app, msg])
    thr.start()
    return thr

